from pbxproj.pbxextensions import TreeType
from pbxproj.pbxsections import PBXGroup, PBXFileReference


class ProjectGroups:
    """
    This class provides separation of concerns, this class contains all methods related to groups manipulations.
    This class should not be instantiated on its own
    """

    def __init__(self):
        raise EnvironmentError('This class cannot be instantiated directly, use XcodeProject instead')

    def add_group(self, name, path=None, parent=None, source_tree='<group>'):
        """
        Add a new group with the given name and optionally path to the parent group. If parent is None, the group will
        be added to the 'root' group.
        Additionally the source tree type can be specified, normally it's group.
        :param name: Name of the group to be added (visible folder name)
        :param path: Path relative to the project where this group points to. If not present, name will match the path
            name
        :param parent: The PBXGroup that will be the parent of this group. If parent is None, the default 'root' group
            will be used as parent
        :param source_tree: The type of group to be created
        :return: PBXGroup created
        """
        group = PBXGroup.create(name=name, path=path, tree=source_tree)

        parent = self._get_parent_group(parent)

        parent.add_child(group)
        self.objects[group.get_id()] = group

        return group

    def remove_group_by_id(self, group_id, recursive=True):
        """
        Remove the group matching the given group_id. If recursive is True, all descendants of this group are also removed.
        :param group_id: The group id to be removed
        :param recursive: All descendants should be removed as well
        :return: True if the element was removed successfully, False if an error occured or there was nothing to remove.
        """
        group = self.objects[group_id]
        if group is None:
            return False

        result = True
        # iterate over the children and determine if they are file/group and call the right method.
        for subgroup_id in list(group.children):
            subgroup = self.objects[subgroup_id]
            if subgroup is None:
                return False

            if recursive and isinstance(subgroup, PBXGroup):
                result &= self.remove_group_by_id(subgroup.get_id(), recursive)
            if isinstance(subgroup, PBXFileReference):
                result &= self.remove_file_by_id(subgroup.get_id())

        if not result:
            return False

        del self.objects[group.get_id()]

        # remove the reference from any other group object that could be containing it.
        for other_group in self.objects.get_objects_in_section('PBXGroup'):
            other_group.remove_child(group)

        return True

    def remove_group_by_name(self, group_name, recursive=True):
        """
        Remove the groups matching the given name. If recursive is True, all descendants of this group are also removed.
        This method could cause the removal of multiple groups that not necessarily are intended to be removed, use with
        caution
        :param group_name: The group name to be removed
        :param recursive: All descendants should be removed as well
        :return: True if the element was removed successfully, False otherwise
        """
        groups = self.get_groups_by_name(name=group_name)

        if groups.__len__() == 0:
            return False

        for group in groups:
            if not self.remove_group_by_id(group.get_id(), recursive):
                return False

        return True

    def get_groups_by_name(self, name, parent=None):
        """
        Retrieve all groups matching the given name and optionally filtered by the given parent node.
        :param name: The name of the group that has to be returned
        :param parent: A PBXGroup object where the object has to be retrieved from. If None all matching groups are returned
        :return: An list of all matching groups
        """
        groups = self.objects.get_objects_in_section('PBXGroup')
        groups = [group for group in groups if group.get_name() == name]

        if parent:
            return [group for group in groups if parent.has_child(group)]

        return groups

    def get_groups_by_path(self, path, parent=None):
        """
        Retrieve all groups matching the given path and optionally filtered by the given parent node.
        The path is converted into the absolute path of the OS before comparison.
        :param path: The name of the group that has to be returned
        :param parent: A PBXGroup object where the object has to be retrieved from. If None all matching groups are returned
        :return: An list of all matching groups
        """
        groups = self.objects.get_objects_in_section('PBXGroup')
        groups = [group for group in groups if group.get_path() == path]

        if parent:
            return [group for group in groups if parent.has_child(group)]

        return groups

    def get_or_create_group(self, name, path=None, parent=None, make_relative=False):
        if not name:
            return None

        groups = self.get_groups_by_name(name, parent)
        if groups.__len__() > 0:
            return groups[0]

        if make_relative and path is not None:
            _, path, _ = self._get_path_and_tree(self._source_root, path, TreeType.SOURCE_ROOT)

        return self.add_group(name, path, parent)

    def _get_parent_group(self, parent):
        if parent is None:
            # search for the mainGroup of the project
            project = self.objects[self['rootObject']]
            if project:
                parent = self.objects[project['mainGroup']]
                if parent is not None:
                    return parent

            # search for the group without name
            parent = self.get_groups_by_name(None)

            # if there is no parent, create and empty parent group, add it to the objects
            if parent.__len__() > 0:
                return parent[0]

            parent = PBXGroup.create(path=None, name=None)
            self.objects[parent.get_id()] = parent
            return parent

        # it's not a group instance, assume it's an id
        if not isinstance(parent, PBXGroup):
            return self.objects[parent]

        return parent



from pbxproj import PBXGenericObject


class XCLocalSwiftPackageReference(PBXGenericObject):
    @classmethod
    def create(cls, relative_path):
        return cls().parse({
            '_id': cls._generate_id(),
            'isa': cls.__name__,
            'relativePath': relative_path
        })

    def _get_comment(self):
        return f'XCLocalSwiftPackageReference "{self.relativePath}"'



"""
usage:
    pbxproj show [options] <project>
    pbxproj show [options] (--target <target>...) <project> [(-s | --source-files) |
                                                             (-H | --header-files) |
                                                             (-r | --resource-files) |
                                                             (-f | --framework-files) |
                                                             (--build-phase-files <build_phase_type>)]

positional arguments:
    <project>                      Project path to the .xcodeproj folder.

generic options:
    -h, --help                      This message.
    -t, --target <target>           Target name to be modified. If there is no target specified, all targets are used.
    -b, --backup                    Creates a backup before start processing the command.

target options:
    -s, --source-files              Show the source files attached to the target
    -r, --resource-files            Show the resource files attached to the target
    -f, --framework-files           Show the library files attached to the target
    -H, --header-files              Show the header files attached to the target
    -c, --configurations            Show the configurations attached to the target
    -B, --build-phase-files <type>  Show the files associated to the build phase of the given type.
"""


def execute(project, args):
    # make a decision of what function to call based on the -D flag
    if args['--target']:
        return _target_info(project, args['--target'], args)
    else:
        return _summary(project, args)


def _summary(project, _):
    info = ''
    for target in project.objects.get_targets():
        configs = ', '.join([c.name for c in project.objects.get_configurations_on_targets(target.name)])
        info += f"{target.name}:\n" \
                f"\tTarget type: {target.isa}\n" \
                f"\tProduct name: {target.productName}\n" \
                f"\tConfigurations: {configs}\n" \

        for build_phase_id in target.buildPhases:
            build_phase = project.objects[build_phase_id]
            info += f"\t{build_phase._get_comment()} ({build_phase.isa}) file count: {build_phase.files.__len__()}\n"

        info += "\n"
    return info


def _target_info(project, target_name, args):
    build_phases = _process_parameters(args)

    info = ''
    for target in project.objects.get_targets(target_name):
        info += f"{target.name}:\n" \
                f"\tProduct name: {target.productName}\n"

        if args['--configurations']:
            configs = ', '.join([c.name for c in project.objects.get_configurations_on_targets(target.name)])
            info += f"\tConfigurations: {configs}\n"

        for build_phase_id in target.buildPhases:
            build_phase = project.objects[build_phase_id]
            if build_phase.isa in build_phases:
                info += f"\t{build_phase._get_comment()}: \n\t\t"
                files = []
                for build_file_id in build_phase.files:
                    build_file = project.objects[build_file_id]
                    files.append(project.objects[build_file.fileRef]._get_comment())
                formatted_files = "\n\t\t".join(sorted(files))
                info += f'{formatted_files}\n'
        info += '\n'
    return info


def _process_parameters(args):
    build_phases = []
    if args['--source-files']:
        build_phases += ['PBXSourcesBuildPhase']
    elif args['--header-files']:
        build_phases += ['PBXHeadersBuildPhase']
    elif args['--resource-files']:
        build_phases += ['PBXResourcesBuildPhase']
    elif args['--framework-files']:
        build_phases += ['PBXFrameworksBuildPhase']
    elif args['--build-phase-files']:
        build_phases += [args['--build-phase-files']]
    return build_phases



class PBXKey(str):
    def __new__(cls, value, parent):
        obj = str.__new__(cls, value)
        obj._parent = parent
        return obj

    def __repr__(self):
        comment = self._get_comment()
        if comment is not None:
            comment = f' /* {comment} */'
        else:
            comment = ''

        return f'{self.__str__()}{comment}'

    def get_parent(self):
        return self._parent

    def _get_comment(self):
        return self.get_parent()._resolve_comment(self)



import os

from pbxproj.XcodeProject import XcodeProject

PROJECT_PLACEHOLDER = '<project>'
PATH_PLACEHOLDER = '<path>'


def open_project(args):
    if os.path.isdir(args[PROJECT_PLACEHOLDER]):
        args[PROJECT_PLACEHOLDER] += '/project.pbxproj'

    if not os.path.isfile(args[PROJECT_PLACEHOLDER]):
        raise FileNotFoundError('Project file not found')

    return XcodeProject.load(args[PROJECT_PLACEHOLDER])


def backup_project(project, args):
    if args['--backup']:
        return project.backup()
    return None


def resolve_backup(project, backup_file, args):
    # remove backup if everything was ok.
    if args['--backup'] and backup_file:
        os.remove(backup_file)


def command_parser(command, auto_save=True):
    def parser(args):
        try:
            project = open_project(args)
            backup_file = backup_project(project, args)
            print((command(project, args)))
            if auto_save:
                project.save()
            resolve_backup(project, backup_file, args)
        except Exception as ex:
            print(f'{ex}')
            exit(1)
    return parser



from pbxproj import PBXGenericObject


class XCRemoteSwiftPackageReference(PBXGenericObject):
    @classmethod
    def create(cls, repository_url, requirement):
        return cls().parse({
            '_id': cls._generate_id(),
            'isa': cls.__name__,
            'repositoryURL': repository_url,
            'requirement': PBXGenericObject().parse(requirement)
        })

    def _get_comment(self):
        name = None
        if hasattr(self, 'repositoryURL'):
            name = self.repositoryURL.split('/')[-1]
            if name.endswith('.git'):
                name = name[:-4]
        return f'XCRemoteSwiftPackageReference "{name}"'



from pbxproj.pbxsections.PBXBuildFile import *
from pbxproj.pbxsections.PBXFileReference import *
from pbxproj.pbxsections.PBXFrameworksBuildPhase import *
from pbxproj.pbxsections.PBXProject import *
from pbxproj.pbxsections.PBXResourcesBuildPhase import *
from pbxproj.pbxsections.PBXSourcesBuildPhase import *
from pbxproj.pbxsections.XCConfigurationList import *
from pbxproj.pbxsections.PBXContainerItemProxy import *
from pbxproj.pbxsections.PBXCopyFilesBuildPhase import *
from pbxproj.pbxsections.PBXShellScriptBuildPhase import *
from pbxproj.pbxsections.PBXTargetDependency import *
from pbxproj.pbxsections.PBXHeadersBuildPhase import *
from pbxproj.pbxsections.XCBuildConfiguration import *
from pbxproj.pbxsections.PBXAggregateTarget import *
from pbxproj.pbxsections.PBXNativeTarget import *
from pbxproj.pbxsections.PBXLegacyTarget import *
from pbxproj.pbxsections.PBXReferenceProxy import *
from pbxproj.pbxsections.PBXGroup import *
from pbxproj.pbxsections.XCSwiftPackageProductDependency import *
from pbxproj.pbxsections.XCRemoteSwiftPackageReference import *
from pbxproj.pbxsections.XCLocalSwiftPackageReference import *



from pbxproj import PBXGenericObject


class XCSwiftPackageProductDependency(PBXGenericObject):
    @classmethod
    def create(cls, package_ref, product_name):
        return cls().parse({
            '_id': cls._generate_id(),
            'isa': cls.__name__,
            'package': package_ref.get_id(),
            'productName': product_name
        })

    def _get_comment(self):
        if hasattr(self, 'productName'):
            return self.productName

        return None



import os
from collections import OrderedDict

from pbxproj import PBXGenericObject


class XCBuildConfigurationFlags:
    OTHER_CFLAGS = 'OTHER_CFLAGS'
    OTHER_LDFLAGS = 'OTHER_LDFLAGS'
    HEADER_SEARCH_PATHS = 'HEADER_SEARCH_PATHS'
    LIBRARY_SEARCH_PATHS = 'LIBRARY_SEARCH_PATHS'
    FRAMEWORK_SEARCH_PATHS = 'FRAMEWORK_SEARCH_PATHS'
    LD_RUNPATH_SEARCH_PATHS = 'LD_RUNPATH_SEARCH_PATHS'


class XCBuildConfiguration(PBXGenericObject):

    def add_flags(self, flag_name, flags):
        if 'buildSettings' not in self:
            self['buildSettings'] = PBXGenericObject()

        current_flags = self.buildSettings[flag_name]
        if current_flags is None:
            self.set_flags(flag_name, flags)
            return

        if not isinstance(current_flags, list):
            current_flags = [current_flags]

        if not isinstance(flags, list):
            flags = [flags]

        self.set_flags(flag_name, current_flags + flags)

    def set_flags(self, flag_name, flags):
        if 'buildSettings' not in self:
            self['buildSettings'] = PBXGenericObject()

        if not isinstance(flags, list):
            flags = [flags]

        self.buildSettings[flag_name] = list(OrderedDict.fromkeys(flags))

    def remove_flags(self, flag_name, flags):
        if 'buildSettings' not in self or self.buildSettings[flag_name] is None:
            # nothing to remove
            return False

        current_flags = self.buildSettings[flag_name]
        if not isinstance(current_flags, list):
            current_flags = [current_flags]

        if flags is None:
            flags = current_flags

        if not isinstance(flags, list):
            flags = [flags]

        self.buildSettings[flag_name] = [x for x in current_flags if x not in flags]
        return True

    def add_search_paths(self, key, paths, recursive=False, escape=False):
        if not isinstance(paths, list):
            paths = [paths]

        # build the recursive/escaped strings and add the flags accordingly
        flags = []
        for path in paths:
            if path == '$(inherited)':
                escape = False
                recursive = False

            if escape:
                path = f'"{path}"'

            if recursive and not path.endswith('/**'):
                path = os.path.join(path, '**')

            flags.append(path)

        self.add_flags(key, flags)

    def remove_search_paths(self, key, paths):
        return self.remove_flags(key, paths)



from pbxproj import PBXGenericObject


class XCConfigurationList(PBXGenericObject):
    def _get_comment(self):
        info = self._get_section()
        return f'Build configuration list for {info[0]} "{info[1]}"'

    def _get_section(self):
        objects = self.get_parent()
        target_id = self.get_id()

        for obj in objects.get_objects_in_section('PBXNativeTarget', 'PBXLegacyTarget', 'PBXAggregateTarget'):
            if target_id in obj.buildConfigurationList:
                return obj.isa, obj.name

        projects = filter(lambda o: target_id in o.buildConfigurationList, objects.get_objects_in_section('PBXProject'))
        project = projects.__next__()
        target = objects[project.targets[0]]
        name = target.name if hasattr(target, 'name') else target.productName
        return project.isa, name
